from .gui import Logger, Oscilloscope
from .options import MySettings, Output_Signal_Settings ,set_plot_colours
from .file import load_data, save_data, save_fig, export_to_matlab_jwlogger, export_to_matlab, export_to_csv
# from .oscilloscope import Oscilloscope
from .acquisition import log_data, output_signal, signal_generator, stream_snapshot
from .datastructure import DataSet, TimeData, FreqData, CrossSpecData, TfData, SonoData, MetaData, ModalData, update_dataset
from .testdata import create_test_impulse_data, create_test_impulse_ensemble, create_test_noise_data
from .plotting import PlotData
from .analysis import calculate_fft, calculate_cross_spectrum_matrix, calculate_cross_spectra_averaged, clean_impulse
from .analysis import calculate_tf, calculate_tf_averaged, multiply_by_power_of_iw, best_match, calculate_sonogram
from .streams import Recorder, Recorder_NI, start_stream, REC, setup_output_NI, setup_output_soundcard, list_available_devices, get_devices_NI, get_devices_soundcard
from .modal import modal_fit_single_channel, modal_fit_all_channels
# import faulthandler
# faulthandler.enable()
# from .gui_tk_test import Logger




# tkinter import common modules
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox

# tkinter matplotlib backend
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
from matplotlib.ticker import AutoLocator

import copy
import numpy as np
import threading




# #import logging
# #logging.basicConfig(filename='example.log',level=logging.DEBUG)
# #%%

# from . import plotting
# from . import datastructure
# from . import acquisition
# from . import analysis
# from . import streams
# from . import file
# from . import modal
# from . import options
import time
import sys, os

#%%
import tkinter as tk
from tkinter import ttk
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

class Logger(tk.Tk):
    def __init__(self):
        super().__init__()

        # Set up the main window
        self.title("Logger")
        self.geometry("800x600")
        self.iconphoto(True, tk.PhotoImage(file='./icon.png'))
        
        # Create grid of frames
        self.frame_main = ttk.Frame(self)
        self.frame_main.grid(column=0, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        

        # Create left frame
        self.left_frame = ttk.Frame(self.frame_main)
        self.left_frame.grid(column=0, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        
        # Create right frame
        self.right_frame = ttk.Frame(self.frame_main)
        self.right_frame.grid(column=2, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        
        
        # Create middle frame
        self.middle_frame = ttk.Frame(self.frame_main)
        self.middle_frame.grid(column=1, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        
        
        # Create top middle frame
        self.top_middle_frame = ttk.Frame(self.middle_frame)
        self.top_middle_frame.grid(column=0, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        
        # Create centre middle frame
        self.centre_middle_frame = ttk.Frame(self.middle_frame)
        self.centre_middle_frame.grid(column=0, row=1, sticky=(tk.N, tk.W, tk.E, tk.S))
        
        
        # Create bottom middle frame
        self.bottom_middle_frame = ttk.Frame(self.middle_frame)
        self.bottom_middle_frame.grid(column=0, row=1, sticky=(tk.N, tk.W, tk.E, tk.S))
        
        

        self.setup_left_frame()
        self.mainloop()

    def setup_left_frame(self):
        # add buttons
        self.button_start = ttk.Button(self.left_frame, text="Start")
        self.button_start.grid(column=0, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        
        
    
    def show(self):
        self.mainloop()
        self.lift()


# %%



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 27 13:28:39 2018

@author: tb267
"""

from . import streams

import numpy as np

import pyqtgraph as pg
import matplotlib.pyplot as plt
import seaborn as sns

# try:
#     import pyaudio
# except ImportError:
#     pyaudio = None
# except NotImplementedError:
#     pyaudio = None


try:
    import sounddevice as sd
except ImportError:
    sd = None
except NotImplementedError:
    sd = None

class MySettings(object):
    '''
    A class that stores the acquisition settings.
    
    Attributes:
    --------------       
    channels: int
        Number of Channels
    fs: int
        Sampling Frequency
    nbits: int
        Number of bits - either 8, 16, 24 or 32
    chunk_size: int
        Number of samples obtained from each channel in one chunk
    num_chunks: int
        Number of chunks to store in circular buffer
    view_time: float
        If specified, overrides num_chunks to display view_time in seconds for oscilloscope
    stored_time: float
        Length of the pre-trigger when the space button is hit, in seconds
    device_index: int
        device index, will prompt if not specified
    init_view_time: bool
        flag for time domain view in oscilloscope
    init_view_freq: bool
        flag for frequency domain view in oscilloscope
    init_view_levels: bool
        flag for channel levels view in oscilloscope
    '''
        
    def __init__(self, *, 
                 channels=2, 
                 fs=44100, 
                 nbits=16, 
                 chunk_size=100, 
                 num_chunks=6,
                 viewed_time=0.3,
                 stored_time=2,
                 pretrig_samples=None,
                 pretrig_threshold=0.05,
                 pretrig_channel=0,
                 pretrig_timeout=20,
                 device_driver='soundcard',
                 device_index=None,
                 VmaxNI=5,
                 NI_mode='DAQmx_Val_RSE',
                 init_view_time=True,
                 init_view_freq=True,
                 init_view_levels=True,
                 output_device_driver=None,
                 output_device_index=None,
                 output_channels=None,
                 output_fs=None,
                 use_output_as_ch0=False):
        
        #INPUT SETTINGS
        self.fs=int(fs)
        self.channels=int(channels)
        self.stored_time=float(stored_time)
        
        if (pretrig_samples is None) or (pretrig_samples == 'None'):
            self.pretrig_samples=None
        else:
            self.pretrig_samples=int(pretrig_samples)
        
        self.pretrig_threshold=float(pretrig_threshold)
        self.pretrig_channel=int(pretrig_channel)
        self.pretrig_timeout=float(pretrig_timeout)
        
        self.device_driver=device_driver
        self.device_index=device_index
        
        
        
        #OUTPUT SETTINGS
        if (output_fs is None) or (output_fs == 'None'):
            self.output_fs = int(self.fs)
        else:
            self.output_fs = int(output_fs)
            
        if (output_channels is None) or (output_channels == 'None'):
            self.output_channels = int(1)
        else:
            self.output_channels = int(output_channels)
            
        # if output device driver not specified then use same as input device
        if (output_device_driver == None) or (output_device_driver == 'None'):
            output_device_driver = device_driver

        self.output_device_driver = output_device_driver
            
        if (device_driver == 'soundcard') and ((device_index == None) or (device_index == 'None')):
            try:
                # try to find default input soundcard device
                self.device_index = sd.default.device[0]
            except:
                # if info not available, select index 1
                self.device_index = 1
        elif (device_driver == 'nidaq') and ((device_index == None) or (device_index == 'None')):
            self.device_index = 0
        else:
            self.device_index = int(device_index)

       
                
        # set output device index to defaults if not specified
        if (output_device_driver == 'soundcard') and ((output_device_index == None) or  (output_device_index == 'None')):
            try:
                # try to find default output soundcard device
                self.output_device_index = sd.default.device[1]
            except:
                # try to guess sensible default output soundcard by string matching device names
                devices = streams.get_devices_soundcard()
                if devices != None:
                    output_devices = np.where(['output' in names for names in devices])
                    self.output_device_index = output_devices[0][0]
                else:
                    self.output_device_index = 1
        elif (output_device_driver == 'nidaq') and ((output_device_index == None) or (output_device_index == 'None')):
            self.output_device_index = 0
        else:
            self.output_device_index = int(output_device_index)
        
        self.use_output_as_ch0 = bool(use_output_as_ch0)

        
        # ADVANCED SETTINGS
        self.VmaxNI=float(VmaxNI)
        self.NI_mode=NI_mode
        self.chunk_size=int(chunk_size)
        self.num_chunks=int(num_chunks)
        self.viewed_time=float(viewed_time)
        self.nbits=int(nbits)
        self.init_view_time=bool(init_view_time)
        self.init_view_freq=bool(init_view_freq)
        self.init_view_levels=bool(init_view_levels)

        ### derived settings
        if (viewed_time != None) and (viewed_time != 'None'):
            self.viewed_time = float(viewed_time)
            self.num_chunks = int(np.ceil(self.viewed_time*self.fs/self.chunk_size))
        else:
            self.viewed_time = None
        
        if self.chunk_size < 10:
            self.chunk_size = int(10)
            print('Resetting ''chunk_size'' to minimum value of 10')
            
        try:    
            self.format = eval('int'+str(self.nbits))
        except:
            pass
        
        self.device_name = None # until initialise stream
        
        if (pretrig_samples == None) or (pretrig_samples == 'None'):
            self.pretrig_samples = None
        else:
            self.pretrig_samples = int(pretrig_samples)
            if self.pretrig_samples > self.chunk_size:
                raise Exception('pretrig_samples must be less than or equal to chunk_size (chunk_size={}).'.format(self.chunk_size))
        
            
        
    
    def __repr__(self):
        template = "{:>24}: {}" # column widths: 8, 10, 15, 7, 10
        #print template.format("CLASSID", "DEPT", "COURSE NUMBER", "AREA", "TITLE") # header
        settings_dict = self.__dict__
        text = '\n<MySettings class>\n\n'
        for attr in settings_dict: 
            text += template.format(attr,settings_dict[attr])
            text += '\n'
        
        return text

        


class Output_Signal_Settings(object):
    '''
    A class that stores the output signal settings.
    '''
    def __init__(self,output_signal_type='None',
                 output_amp = 0,
                 output_f1 = 0,
                 output_f2 = 0):
        self.output_signal_type = output_signal_type
        self.output_amp = output_amp
        self.output_f1 = output_f1
        self.output_f2 = output_f2
        
    

def set_plot_colours(channels):
    '''
    Returns a list of RGB colours depending on the number of channels required.
    '''
    #TODO: Accessible colours
    if channels <= 1:
        cmap = plt.get_cmap('tab10')
#        cmap = sns.color_palette('paired')
        c_list = np.array((np.array(cmap.colors) * 255),dtype=int)
    else:
#        cmap = plt.get_cmap('plasma')
#        v = np.linspace(0,1,channels)
#        c_list = np.array(cmap(v) * 255,dtype=int)
        cmap = sns.hls_palette(channels, l=.3, s=1)
        c_list = np.array(np.array(cmap) * 255,dtype=int)
    
#    val = [0.0,0.5,1.0]
#    colour = np.array([[255,0,0,255],[0,255,0,255],[0,0,255,255]], dtype = np.ubyte)
#    plot_colourmap =  pg.ColorMap(val,colour)
#    c_list = plot_colourmap.getLookupTable(nPts =channels,alpha=True)
    return c_list 



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 27 17:08:42 2018

@author: tb267
"""

from . import datastructure
from . import streams

import numpy as np
import scipy.signal as signal
import scipy.stats as stats
import datetime
import time

MESSAGE = ''

#%% Main data acquisition function
def log_data(settings,test_name=None,rec=None, output=None):
    '''
    Logs data according to settings and returns DataSet class
    '''
    global MESSAGE
    
    if rec is None:
        streams.start_stream(settings)
        rec = streams.REC
        
    
    streams.REC.trigger_detected = False
    
    # Stream is slightly longer than settings.stored_time, so need to add delay
    # from initialisation to allow stream to fill up and prevent zeros at start
    # of logged data.
    time.sleep(2*settings.chunk_size/settings.fs)
    t = datetime.datetime.now()
    timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
    
    if settings.pretrig_samples == None:

        MESSAGE = 'Logging data for {} seconds.\n'.format(settings.stored_time)
        print(MESSAGE)
        
        
        # basic way to control logging time: won't be precise time from calling function
        # also won't be exactly synced to output signal
        if output is not None:
            s = output_signal(settings,output)
            # rec.write(output.astype('float32'))
        
        if (settings.device_driver == 'nidaq') or (output is None): # nidaq output is nonblocking, but soundcard is blocking
            time.sleep(settings.stored_time)
        
            
        # make copy of data
        stored_time_data_copy = np.copy(streams.REC.stored_time_data)
        number_samples = int(streams.REC.settings.stored_time * streams.REC.settings.fs)
        
        stored_time_data_copy = stored_time_data_copy[-number_samples:,:]
        MESSAGE += 'Logging complete.\n'
        print(MESSAGE)

        if output is not None:
            if settings.output_device_driver == 'soundcard':
                s.stop()
                s.close()
            if settings.output_device_driver == 'nidaq':
                s.WaitUntilTaskDone(settings.stored_time+5)
                s.StopTask()
        

        
    else:
        streams.REC.__init__(settings) # zeros buffers so looks for trigger in fresh data
        streams.REC.trigger_detected = False # somehow this can be true even after previous call
        streams.REC.trigger_first_detected_message = True
        
        MESSAGE = 'Waiting for trigger on channel {}.\n'.format(settings.pretrig_channel)
        print(MESSAGE)
        
        start_output_flag = True
        t0 = time.time()
        while (time.time()-t0 < settings.pretrig_timeout) and not streams.REC.trigger_detected:
            if (output is not None) and (start_output_flag == True): # start output within loop, but only once!
                time.sleep(1)
                MESSAGE = 'Starting output signal.\n'
                print(MESSAGE)
                s = output_signal(settings,output)
                # rec.write(output.astype('float32'))
                start_output_flag = False

            time.sleep(0.2)
        if (time.time()-t0 > settings.pretrig_timeout):
            MESSAGE = 'Trigger not detected within timeout of {} seconds.\n'.format(settings.pretrig_timeout)
            print(MESSAGE)
        
        # make copy of data
        stored_time_data_copy = np.copy(streams.REC.stored_time_data)
        trigger_check = stored_time_data_copy[(settings.chunk_size):(2*settings.chunk_size),settings.pretrig_channel]
        detected_sample = settings.chunk_size + np.where(np.abs(trigger_check) > settings.pretrig_threshold)[0][0]
        number_samples = int(settings.stored_time * settings.fs)
        start_index = detected_sample - settings.pretrig_samples
        end_index   = start_index + number_samples
        streams.REC.trigger_detected = False # don't start stream again until sorted out trigger detection
        
        stored_time_data_copy = stored_time_data_copy[start_index:end_index,:]
        
        MESSAGE = 'Logging complete.\n'
        print(MESSAGE)

        if output is not None:
            if settings.output_device_driver == 'soundcard':
                s.stop()
                s.close()
            if settings.output_device_driver == 'nidaq':
                s.WaitUntilTaskDone(settings.stored_time+5)
                s.StopTask()
        
    # make into dataset
    fs = settings.fs
    n_samp = len(stored_time_data_copy[:,0])
    dt = 1/fs
    t_samp = n_samp*dt
    time_axis = np.linspace(0,(n_samp-1)/n_samp * settings.stored_time,n_samp)
    
    if (output is not None) and (settings.use_output_as_ch0 == True):
        stored_output = np.zeros((n_samp,len(output[0,:])))
        n_start = settings.pretrig_samples
        if n_start is None:
            n_start = 0
        n_end = np.copy(n_samp)
        if len(output[:,0]) >= (n_end-n_start):
            stored_output[n_start:n_end,:] = output[:n_end-n_start,:]
        elif len(output[:,0]) < (n_end-n_start):
            stored_output[n_start:len(output[:,0])+n_start,:] = output[:,:]
            
        stored_time_data_copy = np.concatenate((stored_output,stored_time_data_copy),axis=1)
    
    timedata = datastructure.TimeData(time_axis,stored_time_data_copy,settings,timestamp=t,timestring=timestring,test_name=test_name)
    
    
    dataset  = datastructure.DataSet()
    dataset.add_to_dataset(timedata)
    
    # check for clipping
    if np.any(np.abs(stored_time_data_copy) > 0.95):
        MESSAGE += 'WARNING: Data may be clipped'
        print(MESSAGE)
    
    
    return dataset
    


#def log_data_with_output(settings, output,test_name=None, rec=None):
#
    
    # # call log_data function
    # dataset = log_data(settings, test_name, rec)
    
    # # call output_signal function
    # output_signal(settings,output)
    
    # return dataset
    


def output_signal(settings,output):
    # setup NI / audio stream
    if settings.output_device_driver == 'soundcard':
        s = streams.setup_output_soundcard(settings)
        data = output.astype('float32')
        s.write(data)
        return s
        
    elif settings.output_device_driver == 'nidaq':
        sh = np.shape(output)
        T = sh[0]/settings.fs
        s = streams.setup_output_NI(settings,output)
        s.StartTask()
        return s
    else:
        print('device_driver not recognised')
        return None
        
    # send to device


def signal_generator(settings,sig='gaussian',T=1,amplitude=0.1,f=None,selected_channels='all'):
    """
    Creates a signal ready for output to a chosen device
    """
    global MESSAGE
    if selected_channels == 'all':
        selected_channels = np.arange(0,settings.output_channels)
       
    # initiate variables
    t = np.arange(0,T,1/settings.output_fs)
    N_per_channel = np.size(t)
    y = np.zeros((N_per_channel,settings.output_channels))
    win = np.ones((N_per_channel,1))
    T_ramp = np.min([T/10,0.1])
    N_ramp = int(T_ramp*settings.output_fs)
    win[0:N_ramp,0] = 0.5*(1-np.cos(np.arange(0,N_ramp)/N_ramp*np.pi))
    win[-N_ramp:,0] = 0.5*(1+np.cos(np.arange(0,N_ramp)/N_ramp*np.pi))
    
    limit = 1 # generate signals normalised with amplitudes 0-1
    
    # Create sig. Note 'sig' is choice of signal, while 'signal' is scipy.signal
    if sig == 'gaussian':
        y[:,selected_channels] = np.random.randn(N_per_channel,np.size(selected_channels))
        
        
        
#        if settings.output_device_driver == 'soundcard':
#            limit = 1
#        elif settings.output_device_driver == 'nidaq':
#            limit = settings.VmaxNI
             
                
        y[:,selected_channels] = stats.truncnorm.rvs(-limit/amplitude, limit/amplitude, loc=0,scale=amplitude, size=(N_per_channel,np.size(selected_channels)))
        if f is not None:
            b,a = signal.butter(3,f,btype='bandpass',fs=settings.output_fs)
            y = signal.filtfilt(b,a,y,axis=0,padtype=None)
            y = amplitude * y / np.sqrt(np.mean(y**2))
            if np.max(np.abs(y)) > limit:
                y = limit * y / np.max(np.abs(y))
                MESSAGE = 'Actual rms output after scaling to avoid clipping is {0:1.3f}'.format(np.sqrt(np.mean(y**2)))
            else:
                MESSAGE = 'Actual rms output is {0:1.3f}'.format(np.sqrt(np.mean(y**2)))
                
            
#            N_exceed_lim = np.sum(y>limit)+np.sum(y<-limit)
#            fraction_clipped = N_exceed_lim/np.size(y)
#            y[y>limit] = limit
#            y[y<-limit] = -limit
#            if fraction_clipped > 0:
#                print('{} out of {} samples exceeded output voltage limit of device and have been clipped'.format(N_exceed_lim,np.size(y)))
#            if fraction_clipped > 0.01:
#                print('{0:1.1f} percent of samples exceed output voltage limit of device'.format(fraction_clipped*100))
                
    elif sig == 'uniform':
        y[:,selected_channels] = np.random.uniform(low=-amplitude,high=amplitude,size=(N_per_channel,np.size(selected_channels)))
        if f is not None:
            b,a = signal.butter(3,f,btype='bandpass',fs=settings.output_fs)
            y = signal.filtfilt(b,a,y,axis=0,padtype=None)
            y = amplitude * y / np.sqrt(np.mean(y**2))
    elif sig == 'sweep':
        if f is None:
            f = [0,settings.output_fs/2]
        
        for ch in selected_channels:
            y[:,ch] = amplitude*signal.chirp(t,f[0],T,f[1])
    else:
        print('signal type must be one of {''gaussian'',''uniform'',''sweep''}')
        y = np.zeros((N_per_channel,settings.output_channels))
    
    y = win * y
    
    # final correction to ensure all signals limited to 0-1
    if np.max(np.abs(y)) > limit:
        y = limit * y / np.max(np.abs(y))
        MESSAGE = 'Actual rms output after scaling to avoid clipping is {0:1.3f}'.format(np.sqrt(np.mean(y**2)))
        
    return t,y


def stream_snapshot(rec):
    
    time_data_copy = np.copy(streams.REC.osc_time_data)
    time_axis_copy = np.copy(streams.REC.osc_time_axis)
    
    t = datetime.datetime.now()
    timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)    

    time_data = datastructure.TimeData(time_axis_copy,time_data_copy,streams.REC.settings,timestamp=t,timestring=timestring,test_name='stream_snapshot')
    
    
    return time_data